// These 5 lines generate a module that can be included with CommonJS, AMD, and <script> tags.
(function(name, definition) {
    if (typeof module != 'undefined') module.exports = definition()
    else if (typeof define == 'function' && typeof define.amd == 'object') define(definition)
    else this[name] = definition()
}('statebus', function() { var busses = {}, executing_funk, global_funk, funks = {}; return function make_bus () {

    // ****************
    // The statebus object we will return
    function bus (arg) {
        // Called with a function to react to
        if (typeof arg === 'function') {
            var f = reactive(funk)
            f()
            return f
        }

        // Called with a key to produce a subspace with
        return subspace(arg)
    }
    var id = 'bus ' + (Math.random()+'').substring(7)
    bus.toString = function () { return id + (bus.label || '') }
    bus.delete_bus = function () { delete busses[bus.id] }

    // The Data Almighty!!
    var cache = {}

    // ****************
    // Reactive REST API

    var pending_fetches = {}
    var fetches_out = {}                // Maps `key' to `true' iff we've fetched `key'
    var fetches_in = new One_To_Many()  // Maps `key' to `pub_funcs' subscribed to our key
    function fetch (key, callback) {
        key = key.key || key    // You can pass in an object instead of key

        var called_from_reactive_funk = !callback
        var funk = callback || executing_funk

        // Remove this limitation at some point.  One reason for it is
        // that add_handler() doesn't check if a wildcard handler
        // already exists... it just pushes a new one.  That'll grow
        // unbounded.  I can later use regexps for wildcard handlers,
        // and start escaping the patterns between fetch() and
        // add_handler() and solve these issues robustly.
        console.assert(key[key.length-1] !== '*')

        // ** Subscribe the calling funk **
        if (called_from_reactive_funk)
            funk.depends_on(bus, key)
        fetches_in.add(key, funk_key(funk))
        add_handler(key, 'pub', funk)

        // ** Call fetchers upstream **

        // TODO: checking fetches_out[] doesn't count keys that we got
        // which arrived nested within a bigger object, because we
        // never explicity fetched those keys.  But we don't need to
        // fetch them now cause we already have them.
        if (!fetches_out[key])
            bus.route(key, 'fetch', key)

        // ** Return a value **

        // If called reactively, we always return a value.
        if (called_from_reactive_funk)
            return cache[key] = cache[key] || {key: key}

        // Else, we need to pub ONCE.  But each fetch might bein a
        // chain of fetches.  Only the end of the line needs to return
        // a value explicitly from cache, and the others will just
        // forward it along.
        else if (!fetches_out)
            // Then we aren't a link in a chain, and we need to be the
            // one to return the value
            run_handler(funk, 'pub', cache[key] = cache[key] || {key: key})
    }

    function save (obj) { bus.route(obj.key, 'save', obj) }

    function pub (object) {
        delete pending_fetches[object.key]

        var modified_keys = new Set()

        // Recursively add all of object, and its sub-objects, into the cache
        deep_map(object, function (obj) {

            // Two ways to optimize this in future:
            //
            // 1. Only clone objects/arrays if they are new.
            // 
            //    Right now we re-clone all internal arrays and
            //    objects on each pub.  But we really only need to
            //    clone them the first time they are pubbed into the
            //    cache.  After that, we can trust that they aren't
            //    referenced elsewhere.  (We make it the programmer's
            //    responsibility to clone data if necessary on fetch,
            //    but not when on pub.)
            // 
            //    We'll optimize this once we have history.  We can
            //    look at the old version to see if an object/array
            //    existed already before cloning it.
            //
            // 2. Don't go infinitely deep.
            //
            //    Eventually, each save/pub will be limited to the
            //    scope underneath nested keyed objects.  Right now
            //    I'm just recursing infinitely on the whole data
            //    structure with each pub.

            // Clone arrays
            if (Array.isArray(obj))
                obj = obj.slice()

            // Clone objects
            else if (typeof obj === 'object'
                     && obj        // That aren't null
                     && !(obj.key  // That aren't already in cache
                          && cache[obj.key] === obj)) {
                var tmp = {}; for (var k in obj) tmp[k] = obj[k]; obj = tmp
            }

            // Fold cacheable objects into cache
            if (obj && obj.key) {
                modified_keys.add(obj.key)
                if (!cache[obj.key])
                    // This object is new.  Let's store it.
                    cache[obj.key] = obj

                else if (obj !== cache[obj.key]) {
                    // Else, mutate cache to match the object.

                    // First, add/update missing/changed fields to cache
                    for (var k in obj)
                        if (cache[obj.key][k] !== obj[k])
                            cache[obj.key][k] = obj[k]
                    
                    // Then delete extra fields from cache
                    for (var k in cache[obj.key])
                        if (!(k in obj))
                            delete cache[obj.key][k]
                }
                obj = cache[obj.key]
            }

            return obj
        })

        var keys = modified_keys.all()
        for (var i=0; i<keys.length; i++)
            bus.route(keys[i], 'pub', cache[keys[i]])
    }

    function forget (key, pub_handler) {
        pub_handler = pub_handler || global_funk
        var fkey = funk_key(pub_handler)
        //console.log('Fetches in is', fetches_in.hash)
        if (!fetches_in.contains(key, fkey)) {
            console.error("***\n****\nTrying to forget lost key", key,
                          "from function",
                          pub_handler, pub_handler.statebus_id,
                          "that hasn't fetched that key.",
                          funks[fetches_in.get(key)[0]],
                          funks[fetches_in.get(key)[0]] && funks[fetches_in.get(key)[0]].statebus_id
                         )
            throw 'aldkfjalsdfj'
        }

        fetches_in.del(key, fkey)
        del_handler(key, 'pub', pub_handler)

        // If this is the last handler listening to this key, then we
        // can delete the cache entry and send a forget upstream.
        if (!fetches_in.has_any(key)) {
            clearTimeout(to_be_forgotten[key])
            to_be_forgotten[key] = setTimeout(function () {
                bus.route(key, 'forget', key)

                //delete cache[key]
                delete fetches_out[key]
                delete to_be_forgotten[key]
            }, 200)
        }
    }
    function del(obj_or_key) {
        var key = obj_or_key.key || obj_or_key
        delete cache[key]
        console.log('del:', obj_or_key)
        bus.route(key, 'delete', key)
        //forget(key /*, bus??*/)
    }


    // ****************
    // Dirty
    var dirty_keys = new Set()
    var dirty_sweeper = null
    function dirty (key) {
        dirty_keys.add(key)

        dirty_sweeper = dirty_sweeper || setTimeout(function () {
            console.log('dirty_sweeper:', dirty_keys.all())
            // Let's grab the dirty keys and clear it, so that
            // anything that gets dirty during this sweep will be able
            // to sweep again afterward
            var keys = dirty_keys.all()
            dirty_keys.clear()
            dirty_sweeper = null
            
            // Now sweep through our cache of dirty filth and sweep it all up!
            for (var i=0; i<keys.length; i++)
                // If anybody is fetching this key
                if (fetches_in.has_any(keys[i]))
                    bus.route(key, 'fetch', key)

        }, 0)
    }


    // ****************
    // Connections
    function subspace (key) {
        var result = {}
        for (method in {fetch:null, save:null, pub:null,
                        'delete':null, forget:null})
            (function (method) {
                Object.defineProperty(result, 'on_' + method, {
                    set: function (func) { add_handler(key, method, func) },
                    get: function () {
                        var result = handlers_for(key, method)
                        result.remove = function (funk) { del_handler (key, method, funk) }
                        return result
                    }
                })
            })(method)
        return result
    }

    // The funks attached to each key, maps e.g. 'fetch /point/3' to '/30'
    var handlers = new One_To_Many()
    var wildcard_handlers = []  // An array of {prefix, message, funk}
    //var funks = {}              // Maps funk_id -> function

    // A set of timers, for keys to send forgets on
    var to_be_forgotten = {}
    function funk_key (funk) {
        if (!funk.statebus_id) {
            funk.statebus_id = Math.random() + ''
            funks[funk.statebus_id] = funk
        }
        return funk.statebus_id
    }
    function add_handler (key, message, funk) {
        if (key[key.length-1] !== '*')
            handlers.add(message + ' ' + key, funk_key(funk))
        else
            wildcard_handlers.push({prefix: key,
                                    message: message,
                                    funk: funk})

        if (to_be_forgotten[key]) {
            clearTimeout(to_be_forgotten[key])
            delete to_be_forgotten[key]
        }

        // Now check if the method is a fetch and there's a fetched
        // key in this space, and if so call the handler.
    }
    var forget_timer
    function del_handler (key, message, funk) {
        if (key[key.length-1] !== '*')
            // Delete direct connection
            handlers.del(message + ' ' + key, funk_key(funk))
        else
            // Delete wildcard connection
            for (var i=0; i<wildcard_handlers.length; i++) {
                var handler = wildcard_handlers[i]
                if (handler.prefix === key
                    && handler.message === message
                    && handler.funk === funk) {

                    wildcard_handlers.splice(i,1)  // Splice this element out of the array
                    i--                            // And decrement the counter while we're looping
                }
            }
    }

    function handlers_for(key, message) {
        //console.log('handlers_for:', key, message)
        var result = []

        // First get the exact key matches
        var exacts = handlers.get(message + ' ' + key)
        for (var i=0; i < exacts.length; i++)
            result.push(funks[exacts[i]])

        // Now iterate through prefixes
        for (var i=0; i < wildcard_handlers.length; i++) {
            handler = wildcard_handlers[i]

            var prefix = handler.prefix.slice(0, -1)       // Cut off the *
            if (prefix === key.substr(0,prefix.length)     // If the prefix matches
                && message === handler.message)            // And it has the right message
                result.push(handler.funk)
        }

        return result
    }

    function run_handler(funk, method, arg) {
        if (method === 'fetch') {
            fetches_out[arg] = true
            pending_fetches[arg] = funk
        }

        // When we first run a handler (e.g. a fetch or save), we wrap
        // it in a reactive() funk that calls it with its arg.  Then
        // if it fetches or saves, it'll register a pub handler with
        // this funk.

        // Pub events will be calling an already-wrapped funk, that
        // has its own arg
        if (funk.react) {
            console.assert(method === 'pub')
            return funk.react()
        }

        // Fresh fetch/save/forget/delete handlers will just be
        // regular functions.  We'll store their arg and let them
        // re-run until they are done re-running.
        var f = reactive(function () {
            var result = funk(arg)

            // For fetch
            if (method === 'fetch' && result instanceof Object && !f.is_loading()) {
                result.key = arg
                pub(result)
                return result
            }

            // Save aborts changes if still is_loading()
            // ... implement here ...

            // Save, forget and delete handlers stop re-running once
            // they've completed without anything loading.
            // ... with f.forget()
        })

        // on_fetch handlers stop re-running when the key is forgotten
        if (method === 'fetch') {
            var key = arg
            function done () {
                f.forget()
                del_handler(key, 'forget', done)
            }
            add_handler(key, 'forget', done)
        }

        return f()
    }

    // route() can be overridden
    bus.route = function (key, method, arg) {
        var funcs = bus.handlers_for(key, method)
        for (var i=0; i<funcs.length; i++)
            bus.run_handler(funcs[i], method, arg)

        if (method === 'fetch') console.assert(funcs.length<2)
        return funcs.length
    }


    // ****************
    // Reactive functions
    // 
    // We wrap any function with a reactive wrapper that re-calls it
    // whenever state it's fetched changes.

    if (!global_funk) {
        global_funk = reactive(function () {})
        executing_funk = global_funk
        funks[global_funk.statebus_id = 'global'] = global_funk
    }
    //global_funk.fetched_keys = new Set()

    var executing_funk = global_funk
    function reactive(func) {
        var dis, args
        function funk () {
            if (executing_funk !== global_funk)
                console.assert(executing_funk !== funk, 'Recursive funk', args)

            // If you call this function with 
            if (funk.called_directly)
                dis = this, args = arguments

            // Forget the keys from last time
            funk.forget()

            // Now let's run it
            var last_executing_funk = executing_funk
            executing_funk = funk
            try {
                var result = func.apply(dis, args)
            } catch (e) {
                //executing_funk = null // Or should this be last_executing_funk?
                if (funk.is_loading()) return null
                else {
                    console.error('Error!', e)
                    func.apply(dis, args)
                }
            } finally {
                executing_funk = last_executing_funk
            }
            return result
        }

        funk.called_directly = true
        funk.fetched_keys = new One_To_Many() // maps bus to keys
        funk.depends_on = function (bus, key) {
            this.fetched_keys.add(bus, key)
        }
        funk.react = function () {
            var result
            try {
                funk.called_directly = false
                result = funk()
            } finally {
                funk.called_directly = true
            }
            return result
        }
        funk.forget = function () {
            if (funk.statebus_id === 'global') return

            var buss_ids = Object.keys(funk.fetched_keys.hash)
            for (var i=0; i<buss_ids.length; i++) {
                var keys = funk.fetched_keys.get(buss_ids[i])
                for (var j=0; j<keys.length; j++)
                    forget(keys[j], funk)
                funk.fetched_keys.del_all(buss_ids[i])
            }
            // var keys = funk.fetched_keys.all()
            // //console.log('react: forgetting', keys)
            // for (var i=0; i<keys.length; i++) forget(keys[i], funk)
            // funk.fetched_keys.clear()
        }
        funk.is_loading = function () {
            var buss_ids = Object.keys(funk.fetched_keys.hash)
            for (var i=0; i<buss_ids.length; i++) {
                var b = buss_ids[i]
                if (busses[b].loading_keys(funk.fetched_keys.get(b)))
                    return true
            }
            return false
        }
        return funk
    }

    function loading_keys (keys) {
        // Do any of these keys have outstanding gets?
        //console.log('Loading: pending_keys is', pending_fetches)
        for (var i=0; i<keys.length; i++)
            if (pending_fetches[keys[i]]) return true
        return false
    }


    // ******************
    // Utility funcs
    function One_To_Many() {
        var hash = this.hash = {}
        var counts = {}
        this.get = function (k) { return Object.keys(hash[k] || {}) }
        this.add = function (k, v) {
            if (hash[k] === undefined)   hash[k]   = {}
            if (counts[k] === undefined) counts[k] = 0
            if (!hash[k][v]) counts[k]++
            hash[k][v] = true
        }
        this.del = function (k, v) { delete hash[k][v]; counts[k]-- }
        this.del_all = function (k) { delete hash[k]; delete counts[k] }
        this.contains = function (k, v) { return hash[k] && hash[k][v] }
        this.has_any = function (k) { return counts[k] }
    }
    function Set() {
        var hash = {}
        this.add = function (a) { hash[a] = true }
        //this.has = function (a) { return a in hash }
        this.all = function () { return Object.keys(hash) }
        this.del = function (a) { delete hash[a] }
        this.clear = function () { hash = {} }
    }
    function clone(obj) {
        if (obj == null) return obj
        var copy = obj.constructor()
        for (var attr in obj)
            if (obj.hasOwnProperty(attr)) copy[attr] = obj[attr]
        return copy
    }
    function extend(obj, with_obj) {
        if (with_obj === undefined) return obj
        for (var attr in with_obj)
            if (obj.hasOwnProperty(attr)) obj[attr] = with_obj[attr]
        return obj
    }

    function deep_map (object, func) {
        object = func(object)

        // Recurse through each element in arrays
        if (Array.isArray(object))
            for (var i=0; i < object.length; i++)
                object[i] = deep_map(object[i], func)

        // Recurse through each property on objects
        else if (typeof(object) === 'object')
            for (var k in object)
                object[k] = deep_map(object[k], func)

        return object
    }

    function key_id(string) { return string.match(/\/?[^\/]+\/(\d+)/)[1] }
    function key_name(string) { return string.match(/\/?([^\/]+).*/)[1] }


    // #######################################
    // ########### Browser Code ##############
    // #######################################

    // Make the private methods accessible under "window.statebus"
    var api = ['cache fetch save forget del pub dirty',
               'subspace handlers wildcard_handlers handlers_for',
               'run_handler del_handler reactive',
               'funk_key funks key_id key_name id',
               'pending_fetches fetches_in loading_keys',
               'global_funk executing_funk',
               'Set One_To_Many clone extend deep_map'
              ].join(' ').split(' ')
    for (var i=0; i<api.length; i++)
        bus[api[i]] = eval(api[i])

    // Export globals
    if (Object.keys(busses).length === 0) {
        var globals = 'fetch save pub del'.split(' ')
        for (var i=0; i<globals.length; i++)
            this[globals[i]] = /*window[globals[i]] ||*/ eval(globals[i])
    }
    busses[bus.id] = bus
    return bus
}}))
